# -*- coding: utf-8 -*-
import ast
import logging
import symtable
from collections import defaultdict
from types import FrameType
from typing import cast, TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union

import astunparse
import pyccolo as pyc
from IPython import get_ipython

from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs
from nbsafety.data_model.code_cell import cells
from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.data_model.namespace import Namespace
from nbsafety.data_model.scope import Scope
from nbsafety.data_model.timestamp import Timestamp
from nbsafety.run_mode import SafetyRunMode
from nbsafety.singletons import nbs, SingletonBaseTracer
from nbsafety.tracing.mutation_event import (
    ArgMutate,
    ListInsert,
    ListPop,
    ListRemove,
    MutatingMethodEventNotYetImplemented,
    StandardMutation,
    resolve_mutating_method,
)
from nbsafety.tracing.mutation_special_cases import (
    METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN,
    METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN,
)
from nbsafety.tracing.safety_ast_rewriter import SafetyAstRewriter
from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
from nbsafety.tracing.trace_stmt import TraceStatement
from nbsafety.tracing.mutation_event import MutationEvent
from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal_nodes
from nbsafety.types import SupportedIndexType


AttrSubVal = SupportedIndexType
NodeId = int
ObjId = int
MutationCandidate = Tuple[
    Tuple[Any, Optional[str], Optional[str]],
    MutationEvent,
    List[Set[DataSymbol]],
    List[Any],
]
Mutation = Tuple[int, MutationEvent, Set[DataSymbol], List[Any]]
SavedStoreData = Tuple[Namespace, Any, AttrSubVal, bool]
SavedDelData = Tuple[Namespace, Any, AttrSubVal, bool]
SavedComplexSymbolLoadData = Tuple[Namespace, Any, AttrSubVal, bool, Optional[str]]


logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


ARG_MUTATION_EXCEPTED_MODULES = {
    "alt",
    "altair",
    "display",
    "logging",
    "matplotlib",
    "pyplot",
    "plot",
    "plt",
    "seaborn",
    "sns",
    "widget",
}


reactive_spec = pyc.AugmentationSpec(
    aug_type=pyc.AugmentationType.prefix, token="$", replacement=""
)
blocking_spec = pyc.AugmentationSpec(
    aug_type=pyc.AugmentationType.prefix, token="$:", replacement=""
)


class SafetyTracer(SingletonBaseTracer):
    ast_rewriter_cls = SafetyAstRewriter

    def should_propagate_handler_exception(
        self, evt: pyc.TraceEvent, exc: Exception
    ) -> bool:
        return SafetyRunMode.get() == SafetyRunMode.DEVELOP

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracing_enabled_files.discard(self.defined_file)
        with self.persistent_fields():
            self.reactive_node_ids: Set[int] = self.augmented_node_ids_by_spec[
                reactive_spec
            ]
            self.blocking_node_ids: Set[int] = self.augmented_node_ids_by_spec[
                blocking_spec
            ]
        self._module_stmt_counter = 0
        self._saved_stmt_ret_expr: Optional[Any] = None
        self._seen_loop_ids: Set[NodeId] = set()
        self._seen_functions_ids: Set[NodeId] = set()
        self.prev_event: Optional[pyc.TraceEvent] = None
        self.prev_trace_stmt: Optional[TraceStatement] = None
        self.seen_stmts: Set[NodeId] = set()
        self.call_depth = 0
        self.traced_statements: Dict[NodeId, TraceStatement] = {}
        self.node_id_to_loaded_symbols: Dict[NodeId, List[DataSymbol]] = defaultdict(
            list
        )
        self.node_id_to_saved_store_data: Dict[NodeId, SavedStoreData] = {}
        self.node_id_to_saved_live_subscript_refs: Dict[NodeId, Set[DataSymbol]] = {}
        self.node_id_to_saved_del_data: Dict[NodeId, SavedDelData] = {}
        self.node_id_to_loaded_literal_scope: Dict[NodeId, Namespace] = {}
        self.node_id_to_saved_dict_key: Dict[NodeId, Any] = {}
        self.this_stmt_updated_symbols: Set[DataSymbol] = set()
        try:
            self.cur_cell_symtab: symtable.SymbolTable = symtable.symtable(
                cells().current_cell().sanitized_content(),
                f"<cell-{cells().exec_counter()}>",
                "exec",
            )
        except:
            # it'll just give a syntax error anyway when we try to execute;
            # do this just for the benefit of the type checker
            self.cur_cell_symtab: symtable.SymbolTable = symtable.symtable(
                "", f"<cell-{cells().exec_counter()}>", "exec"
            )

        self.call_stack: pyc.TraceStack = self.make_stack()
        with self.call_stack.register_stack_state():
            # everything here should be copyable
            self.prev_trace_stmt_in_cur_frame: Optional[TraceStatement] = None
            self.prev_node_id_in_cur_frame: Optional[NodeId] = None
            self.mutations: List[Mutation] = []
            self.saved_assign_rhs_obj: Optional[Any] = None
            # this one gets set regardless of whether tracing enabled
            self.next_stmt_node_id: Optional[NodeId] = None

            self.pending_class_namespaces: List[Namespace] = []

            with self.call_stack.needing_manual_initialization():
                self.cur_frame_original_scope: Scope = nbs().global_scope
                self.active_scope: Scope = nbs().global_scope
                self.inside_anonymous_call = False

            self.lexical_call_stack: pyc.TraceStack = self.make_stack()
            with self.lexical_call_stack.register_stack_state():
                self.num_args_seen = 0
                self.first_obj_id_in_chain: Optional[ObjId] = None
                self.top_level_node_id_for_chain: Optional[NodeId] = None
                self.saved_complex_symbol_load_data: Optional[
                    SavedComplexSymbolLoadData
                ] = None
                self.prev_node_id_in_cur_frame_lexical: Optional[NodeId] = None
                self.mutation_candidate: Optional[MutationCandidate] = None

                self.lexical_literal_stack: pyc.TraceStack = self.make_stack()
                with self.lexical_literal_stack.register_stack_state():
                    # `None` means use 'cur_frame_original_scope'
                    self.active_literal_scope: Optional[Namespace] = None

    @property
    def syntax_augmentation_specs(self) -> List[pyc.AugmentationSpec]:
        return [blocking_spec, reactive_spec]

    @property
    def should_patch_meta_path(self) -> bool:
        return False

    def module_stmt_counter(self) -> int:
        return self._module_stmt_counter

    # TODO: use stack mechanism to automate this?
    def after_stmt_reset_hook(self) -> None:
        self.mutations.clear()
        self.mutation_candidate = None
        self.active_scope = self.cur_frame_original_scope
        self.first_obj_id_in_chain = None
        self.top_level_node_id_for_chain = None
        self.saved_complex_symbol_load_data = None
        self.active_literal_scope = None
        self.node_id_to_loaded_literal_scope.clear()
        self.node_id_to_saved_dict_key.clear()
        self.prev_node_id_in_cur_frame = None
        self.saved_assign_rhs_obj = None
        nbs().updated_symbols |= self.this_stmt_updated_symbols
        self.this_stmt_updated_symbols.clear()
        self._seen_functions_ids.clear()
        # don't clear the lexical stacks because line magics can
        # mess with when an 'after_stmt' gets emitted, and anyway
        # these should be pushed / popped appropriately by ast events

    def _handle_call_transition(self, trace_stmt: TraceStatement):
        # ensures we only handle del's and not delitem's
        self.node_id_to_saved_del_data.clear()
        new_scope = trace_stmt.get_post_call_scope()
        with self.call_stack.push():
            # TODO: figure out a better way to determine if we're inside a lambda
            #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
            self.inside_anonymous_call = not isinstance(
                trace_stmt.stmt_node,
                (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef),
            )
            self.cur_frame_original_scope = new_scope
            self.active_scope = new_scope
        self.prev_trace_stmt_in_cur_frame = self.prev_trace_stmt = trace_stmt

    def _check_prev_stmt_done_executing_hook(
        self, event: pyc.TraceEvent, trace_stmt: TraceStatement
    ):
        if event == pyc.after_stmt and self.is_tracing_enabled:
            trace_stmt.finished_execution_hook()
        elif event == pyc.return_ and self.prev_event not in (
            pyc.call,
            pyc.exception,
        ):
            # ensuring prev != call ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            if self.prev_trace_stmt is not None:
                self.prev_trace_stmt.finished_execution_hook()
            # prev_overall = self.prev_trace_stmt
            # if prev_overall is not None and prev_overall is not self._stack[-1][0]:
            #     # this condition ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            #     prev_overall.finished_execution_hook()

    def _handle_return_transition(self, trace_stmt: TraceStatement, ret: Any):
        try:
            inside_anonymous_call = self.inside_anonymous_call
            try:
                return_to_stmt: TraceStatement = self.call_stack.get_field(
                    "prev_trace_stmt_in_cur_frame"
                )
            except IndexError:
                # then the first call was triggered from inside library code;
                # skip the transition and disable tracing in case this call
                # happens in a loop; we won't catch it in our normal tracing
                # disabler since it's the first call
                self._disable_tracing()
                return
            assert return_to_stmt is not None
            if self.prev_event != pyc.exception:
                # exception events are followed by return events until we hit an except clause
                # no need to track dependencies in this case
                if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                    return_to_stmt.class_scope = cast(
                        Namespace, self.cur_frame_original_scope
                    )
                elif (
                    isinstance(trace_stmt.stmt_node, ast.Return)
                    or inside_anonymous_call
                ):
                    if not trace_stmt.lambda_call_point_deps_done_once:
                        trace_stmt.lambda_call_point_deps_done_once = True
                        maybe_lambda_sym = nbs().statement_to_func_cell.get(
                            id(trace_stmt.stmt_node), None
                        )
                        maybe_lambda_node = None
                        if maybe_lambda_sym is not None:
                            maybe_lambda_node = maybe_lambda_sym.stmt_node
                        if (
                            inside_anonymous_call
                            and maybe_lambda_node is not None
                            and isinstance(maybe_lambda_node, ast.Lambda)
                        ):
                            rvals = resolve_rval_symbols(maybe_lambda_node.body)
                        else:
                            rvals = resolve_rval_symbols(trace_stmt.stmt_node)
                        dsym_to_attach = None
                        if len(rvals) == 1:
                            dsym_to_attach = next(iter(rvals))
                            if dsym_to_attach.obj_id != id(ret):
                                dsym_to_attach = None
                        if dsym_to_attach is None and len(rvals) > 0:
                            dsym_to_attach = self.cur_frame_original_scope.upsert_data_symbol_for_name(
                                "<return_sym_%d>" % id(ret),
                                ret,
                                rvals,
                                trace_stmt.stmt_node,
                                is_anonymous=True,
                            )
                        if dsym_to_attach is not None:
                            return_to_node_id = self.call_stack.get_field(
                                "prev_node_id_in_cur_frame"
                            )
                            # logger.error("prev seen: %s", ast.dump(self.ast_node_by_id[return_to_node_id]))
                            try:
                                call_node_id = self.call_stack.get_field(
                                    "lexical_call_stack"
                                ).get_field("prev_node_id_in_cur_frame_lexical")
                                call_node = cast(
                                    ast.Call, self.ast_node_by_id[call_node_id]
                                )
                                # logger.error("prev seen outer: %s", ast.dump(self.ast_node_by_id[call_node_id]))
                                total_args = len(call_node.args) + len(
                                    call_node.keywords
                                )
                                num_args_seen = self.call_stack.get_field(
                                    "num_args_seen"
                                )
                                logger.warning("num args seen: %d", num_args_seen)
                                if total_args == num_args_seen:
                                    return_to_node_id = call_node_id
                                else:
                                    assert num_args_seen < total_args
                                    if num_args_seen < len(call_node.args):
                                        return_to_node_id = id(
                                            call_node.args[num_args_seen]
                                        )
                                    else:
                                        return_to_node_id = id(
                                            call_node.keywords[
                                                num_args_seen - len(call_node.args)
                                            ].value
                                        )
                            except IndexError:
                                pass
                            # logger.error("use node %s", ast.dump(self.ast_node_by_id[return_to_node_id]))
                            self.node_id_to_loaded_symbols[return_to_node_id].append(
                                dsym_to_attach
                            )
        finally:
            if self.is_tracing_enabled:
                self.call_stack.pop()
            if nbs().is_develop and len(self.call_stack) == 0:
                assert self.call_depth == 1

    def state_transition_hook(
        self,
        event: pyc.TraceEvent,
        trace_stmt: TraceStatement,
        ret: Any,
    ):
        self._check_prev_stmt_done_executing_hook(event, trace_stmt)

        if event == pyc.call:
            self._handle_call_transition(trace_stmt)
        if event == pyc.return_:
            self._handle_return_transition(trace_stmt, ret)
        self.prev_trace_stmt = trace_stmt
        self.prev_event = event

    @staticmethod
    def _partial_resolve_ref(ref: Union[str, int, ast.AST]) -> Union[str, int]:
        if isinstance(ref, ast.Starred):
            ref = ref.value
        if isinstance(ref, ast.Name):
            ref = ref.id
        if isinstance(ref, ast.AST):
            ref = id(ref)
        return ref

    def _resolve_store_data_for_simple_target(self, target: str, frame: FrameType):
        scope = self.cur_frame_original_scope
        lut = frame.f_locals
        if scope.symtab is not None:
            try:
                target_sym = scope.symtab.lookup(target)
                # this nonsense is necessary because the "is_nonlocal" method
                # is not available on Python <= 3.7;
                # the below check seems to work consistently across all Python versions
                is_nonlocal = getattr(
                    target_sym,
                    "is_nonlocal",
                    lambda: not target_sym.is_global()
                    and target_sym.is_assigned()
                    and target_sym.is_free(),
                )()
                if is_nonlocal:
                    scope = scope.parent_scope
                elif target_sym.is_global():
                    lut = frame.f_globals
                    scope = nbs().global_scope
            except KeyError:
                pass
        try:
            obj = lut[target]
        except KeyError:
            obj = frame.f_globals[target]
            scope = nbs().global_scope
        return scope, target, obj, False, set()

    def resolve_store_data_for_target(
        self, target: Union[str, int, ast.AST], frame: FrameType
    ) -> Tuple[Scope, AttrSubVal, Any, bool, Set[DataSymbol]]:
        target = self._partial_resolve_ref(target)
        if isinstance(target, str):
            return self._resolve_store_data_for_simple_target(target, frame)
        (scope, obj, attr_or_sub, is_subscript) = self.node_id_to_saved_store_data.pop(
            target
        )
        if isinstance(obj, (dict, list)):
            # we can be reasonably sure that the object on the rhs is the same thing
            # that gets stashed in `obj` for these cases, so use it instead of doing
            # the lookup (which may have side effects) to reduce intrusiveness
            attr_or_sub_obj = self.saved_assign_rhs_obj
        else:
            attr_or_sub_obj = nbs().retrieve_namespace_attr_or_sub(
                obj, attr_or_sub, is_subscript
            )
        if attr_or_sub_obj is None:
            scope_to_use = scope
        else:
            scope_to_use = scope.get_earliest_ancestor_containing(
                id(attr_or_sub_obj), is_subscript
            )
        if scope_to_use is None:
            # Nobody before `scope` has it, so we'll insert it at this level
            scope_to_use = scope
        return (
            scope_to_use,
            attr_or_sub,
            attr_or_sub_obj,
            is_subscript,
            self.node_id_to_saved_live_subscript_refs.pop(target, set()),
        )

    def resolve_del_data_for_target(
        self, target: Union[str, int, ast.AST]
    ) -> Tuple[Scope, Optional[Any], AttrSubVal, bool]:
        target = self._partial_resolve_ref(target)
        if isinstance(target, str):
            return self.cur_frame_original_scope, None, target, False
        (scope, obj, attr_or_sub, is_subscript) = self.node_id_to_saved_del_data[target]
        return scope, obj, attr_or_sub, is_subscript

    def resolve_loaded_symbols(
        self, symbol_ref: Union[str, int, ast.AST, DataSymbol]
    ) -> List[DataSymbol]:
        if isinstance(symbol_ref, DataSymbol):
            return [symbol_ref]
        symbol_ref = self._partial_resolve_ref(symbol_ref)
        if isinstance(symbol_ref, int):
            return self.node_id_to_loaded_symbols.get(symbol_ref, [])
        elif isinstance(symbol_ref, str):
            ret = self.cur_frame_original_scope.lookup_data_symbol_by_name(symbol_ref)
            if ret is None:
                return []
            else:
                return [ret]
        else:
            return []

    def resolve_symbols(
        self, symbol_refs: Set[Union[str, int, DataSymbol]]
    ) -> Set[DataSymbol]:
        data_symbols = set()
        for ref in symbol_refs:
            data_symbols.update(self.resolve_loaded_symbols(ref))
        return data_symbols

    def _get_namespace_for_obj(
        self, obj: Any, obj_name: Optional[str] = None
    ) -> Namespace:
        obj_id = id(obj)
        ns = nbs().namespaces.get(obj_id, None)
        if ns is not None:
            return ns
        class_scope = nbs().namespaces.get(id(obj.__class__), None)
        if class_scope is not None:
            # logger.warning(
            #     'found class scope %s containing %s',
            #     class_scope, list(class_scope.all_data_symbols_this_indentation())
            # )
            ns = class_scope.clone(obj)
            if obj_name is not None:
                ns.scope_name = obj_name
        else:
            # print('no scope for class', obj.__class__)
            try:
                scope_name = (
                    nbs().get_first_full_symbol(obj_id).name
                    if obj_name is None
                    else obj_name
                )
            except AttributeError:
                scope_name = "<unknown namespace>"
            ns = Namespace(obj, scope_name, parent_scope=None)
        # FIXME: brittle strategy for determining parent scope of obj
        if ns.parent_scope is None:
            if (
                obj_name is not None
                and obj_name not in self.prev_trace_stmt_in_cur_frame.frame.f_locals
            ):
                parent_scope = nbs().global_scope
            else:
                parent_scope = self.active_scope
            ns.parent_scope = parent_scope
        return ns

    def _clear_info_and_maybe_lookup_or_create_complex_symbol(
        self, obj_attr_or_sub
    ) -> Optional[DataSymbol]:
        if self.saved_complex_symbol_load_data is None:
            return None
        (
            scope,
            obj,
            attr_or_subscript,
            is_subscript,
            *_,
        ) = self.saved_complex_symbol_load_data
        self.saved_complex_symbol_load_data = None
        data_sym = scope.lookup_data_symbol_by_name_this_indentation(
            attr_or_subscript,
            is_subscript=is_subscript,
            skip_cloned_lookup=True,
        )
        logger.warning("found sym %s in scope %s", data_sym, scope)
        if data_sym is None:
            parent = scope.lookup_data_symbol_by_name_this_indentation(
                attr_or_subscript,
                is_subscript,
                skip_cloned_lookup=False,
            )
            parents = set() if parent is None else {parent}
            is_default_dict = isinstance(obj, defaultdict)
            data_sym = scope.upsert_data_symbol_for_name(
                attr_or_subscript,
                obj_attr_or_sub,
                parents,
                self.prev_trace_stmt_in_cur_frame.stmt_node,
                is_subscript=is_subscript,
                propagate=is_default_dict,
                implicit=not is_default_dict,
            )
        elif data_sym.obj_id != id(obj_attr_or_sub):
            data_sym.update_obj_ref(obj_attr_or_sub)
        return data_sym

    @pyc.register_raw_handler(
        (
            pyc.before_call,
            pyc.before_attribute_load,
            pyc.before_attribute_store,
            pyc.before_attribute_del,
            pyc.before_subscript_load,
            pyc.before_subscript_store,
            pyc.before_subscript_del,
        )
    )
    def _save_node_id(self, _obj, node_id: NodeId, frame, *_, **__):
        self.prev_node_id_in_cur_frame = node_id
        self.prev_node_id_in_cur_frame_lexical = node_id

    @pyc.register_raw_handler(pyc.init_module)
    def init_cell(self, _obj, _node_id, frame: FrameType, _event: pyc.TraceEvent, **__):
        nbs().set_name_to_cell_num_mapping(frame)
        self._tracing_enabled_files.add(frame.f_code.co_filename)

    # @pyc.register_raw_handler((pyc.before_for_loop_body, pyc.before_while_loop_body))
    # def before_loop_body(self, _obj: Any, loop_id: NodeId, *_, **__):
    #     ret = self.is_tracing_enabled and loop_id not in self._seen_loop_ids
    #     if ret:
    #         self._seen_loop_ids.add(loop_id)
    #     return ret

    @pyc.register_raw_handler((pyc.after_for_loop_iter, pyc.after_while_loop_iter))
    def after_loop_iter(self, _obj: Any, _loop_id: NodeId, *_, guard: str, **__):
        self.activate_guard(guard)

    # @pyc.register_raw_handler(pyc.after_function_execution)
    # def after_function_exec(self, _obj: Any, _loop_id: NodeId, *_, guard: str, **__):
    #     self.activate_guard(guard)

    @pyc.register_raw_handler(pyc.after_assign_rhs)
    @pyc.skip_when_tracing_disabled
    def after_assign_rhs(self, obj: Any, *_, **__):
        self.saved_assign_rhs_obj = obj

    @pyc.register_raw_handler(pyc.after_subscript_slice)
    @pyc.skip_when_tracing_disabled
    def after_subscript_slice(self, _obj: Any, node_id: NodeId, *__, **___):
        node = self.ast_node_by_id.get(node_id, None)
        if node is None:
            return
        slice_node = cast(ast.Subscript, node).slice
        live, _ = compute_live_dead_symbol_refs(
            slice_node, scope=self.cur_frame_original_scope
        )
        subscript_live_refs = []
        for ref in live:
            if len(ref.ref.chain) == 1:
                subscript_live_refs.append(cast(str, ref.ref.chain[0].value))
        self.node_id_to_saved_live_subscript_refs[node_id] = self.resolve_symbols(
            set(subscript_live_refs)
        )
        Timestamp.update_usage_info(
            self.cur_frame_original_scope.lookup_data_symbol_by_name(ref)
            for ref in subscript_live_refs
        )

    @pyc.register_raw_handler(
        (
            pyc.before_attribute_load,
            pyc.before_attribute_store,
            pyc.before_attribute_del,
            pyc.before_subscript_load,
            pyc.before_subscript_store,
            pyc.before_subscript_del,
        )
    )
    @pyc.skip_when_tracing_disabled
    def attrsub_tracer(
        self,
        obj: Any,
        node_id: NodeId,
        _frame_: FrameType,
        event: pyc.TraceEvent,
        *_,
        attr_or_subscript: AttrSubVal,
        call_context: bool,
        top_level_node_id: NodeId,
        obj_name: Optional[str] = None,
        **__,
    ):
        value_node_id = id(self.ast_node_by_id[node_id].value)  # type: ignore
        if isinstance(self.ast_node_by_id[value_node_id], ast.Call):
            # clear the callpoint dependency
            self.node_id_to_loaded_symbols.pop(value_node_id, None)
        if obj is None or obj is get_ipython():
            return
        logger.warning("%s %s of obj %s", event, attr_or_subscript, obj)
        sym_for_obj = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)

        # Resolve symbol if necessary
        if sym_for_obj is None and obj_name is not None:
            sym_for_obj = self.active_scope.lookup_data_symbol_by_name_this_indentation(
                obj_name
            )

        scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
        is_subscript = "subscript" in event.value
        if sym_for_obj is not None:
            try:
                data_sym = scope.lookup_data_symbol_by_name_this_indentation(
                    attr_or_subscript,
                    is_subscript=is_subscript,
                    skip_cloned_lookup=True,
                )
            except TypeError:
                data_sym = None
            if data_sym is None:
                sym_for_obj.update_usage_info()
            else:
                sym_for_obj.update_usage_info(exclude_ns=True)

        obj_id = id(obj)
        if self.top_level_node_id_for_chain is None:
            self.top_level_node_id_for_chain = top_level_node_id
        if self.first_obj_id_in_chain is None:
            self.first_obj_id_in_chain = obj_id

        try:
            if isinstance(attr_or_subscript, tuple):
                if not all(isinstance(v, (str, int)) for v in attr_or_subscript):
                    return
            elif not isinstance(attr_or_subscript, (str, int)):
                return
            if "store" in event.value:
                logger.warning(
                    "save store data for node id %d: %s, %s, %s, %s",
                    top_level_node_id,
                    scope,
                    obj,
                    attr_or_subscript,
                    is_subscript,
                )
                self.node_id_to_saved_store_data[top_level_node_id] = (
                    scope,
                    obj,
                    attr_or_subscript,
                    is_subscript,
                )
                return
            elif "del" in event.value:
                # logger.error("save del data for node %s", ast.dump(self.ast_node_by_id[top_level_node_id]))
                logger.warning("save del data for node id %d", top_level_node_id)
                self.node_id_to_saved_del_data[top_level_node_id] = (
                    scope,
                    obj,
                    attr_or_subscript,
                    is_subscript,
                )
                return
            logger.warning(
                "saved load data: %s, %s, %s", scope, attr_or_subscript, is_subscript
            )
            self.saved_complex_symbol_load_data = (
                scope,
                obj,
                attr_or_subscript,
                is_subscript,
                obj_name,
            )
            if call_context:
                if not is_subscript:
                    if (
                        sym_for_obj is None
                        and self.prev_trace_stmt_in_cur_frame is not None
                    ):
                        sym_for_obj = self.active_scope.upsert_data_symbol_for_name(
                            obj_name or "<anonymous_symbol_%d>" % id(obj),
                            obj,
                            set(),
                            self.prev_trace_stmt_in_cur_frame.stmt_node,
                            is_subscript=is_subscript,
                            is_anonymous=obj_name is None,
                            propagate=False,
                            implicit=True,
                        )
                    if sym_for_obj is not None:
                        assert self.top_level_node_id_for_chain is not None
                        self.node_id_to_loaded_symbols[
                            self.top_level_node_id_for_chain
                        ].append(sym_for_obj)
        finally:
            self.active_scope = scope

    def _process_possible_mutation(self, retval: Any) -> None:
        if self.mutation_candidate is None:
            return
        (
            (obj, obj_name, method_name),
            mutation_event,
            recorded_arg_dsyms,
            recorded_arg_objs,
        ) = self.mutation_candidate
        self.mutation_candidate = None
        if obj is logging or isinstance(obj, logging.Logger):
            # ignore calls to logging.whatever(...)
            return
        obj_type = None
        obj_id = id(obj)
        if obj_id in nbs().aliases:
            aliases = nbs().aliases[obj_id]
            if len(aliases) > 0:
                obj_type = next(iter(aliases)).obj_type
        if obj_type is None:
            obj_type = type(obj)
        is_excepted_mutation = False
        is_excepted_non_mutation = False
        if isinstance(mutation_event, StandardMutation):
            # only look for exceptions for standard mutations; other cases are handled elsewhere
            if retval is not None and id(retval) != obj_id:
                # doesn't look like something we can trace, but it also
                # doesn't look like something that mutates the caller, since
                # the return value is not None and it's not the caller object
                if (
                    obj_id,
                    method_name,
                ) in METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN:
                    is_excepted_mutation = True
                else:
                    return
            if not is_excepted_mutation:
                if retval is None:
                    is_excepted_non_mutation = (
                        obj_id,
                        method_name,
                    ) in METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN
                if (
                    is_excepted_non_mutation
                    or obj_type is None
                    or id(obj_type) in nbs().aliases
                ):
                    # the calling obj looks like something that we can trace;
                    # no need to process the call as a possible mutation
                    return
        arg_dsyms: Set[DataSymbol] = set()
        arg_dsyms = arg_dsyms.union(*recorded_arg_dsyms)
        if isinstance(mutation_event, StandardMutation):
            try:
                top_level_sym = nbs().get_first_full_symbol(self.first_obj_id_in_chain)
                if (
                    top_level_sym.is_import
                    and top_level_sym.name not in ARG_MUTATION_EXCEPTED_MODULES
                ):
                    # TODO: should it be the other way around?
                    #  i.e. allow-list for arg mutations, starting with np.random.seed?
                    mutated_dsym = None
                    if len(recorded_arg_dsyms) > 0:
                        first_arg_dsyms = list(recorded_arg_dsyms[0])
                        first_arg_dsyms = [
                            dsym
                            for dsym in first_arg_dsyms
                            if dsym.obj is recorded_arg_objs[0]
                        ]
                        if len(first_arg_dsyms) == 1:
                            mutated_dsym = first_arg_dsyms[0]
                            if mutated_dsym.obj_type in DataSymbol.IMMUTABLE_TYPES:
                                mutated_dsym = None
                            elif mutated_dsym.obj_type in {list, set, dict}:
                                # assume module code won't mutate these primitive containers
                                mutated_dsym = None
                    if mutated_dsym is not None:
                        # only make this an arg mutation event if it looks like there's an arg to mutate
                        arg_dsyms = {mutated_dsym}
                        # just consider the first one mutated unless other args depend on it
                        for other_recorded_arg_dsyms in recorded_arg_dsyms[1:]:
                            arg_dsyms.update(
                                {
                                    dsym
                                    for dsym in other_recorded_arg_dsyms
                                    if mutated_dsym in dsym.parents
                                }
                            )
                        mutation_event = ArgMutate()
            except:
                pass
        self.mutations.append((obj_id, mutation_event, arg_dsyms, recorded_arg_objs))

    @pyc.register_raw_handler(pyc.after_load_complex_symbol)
    def after_complex_symbol(self, obj: Any, *_, **__):
        try:
            if not self.is_tracing_enabled:
                return
            if self.first_obj_id_in_chain is None:
                return
            assert self.top_level_node_id_for_chain is not None
            loaded_sym = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
            if loaded_sym is not None:
                self.node_id_to_loaded_symbols[self.top_level_node_id_for_chain].append(
                    loaded_sym
                )
        finally:
            self.saved_complex_symbol_load_data = None
            self.first_obj_id_in_chain = None
            self.top_level_node_id_for_chain = None
            self.active_scope = self.cur_frame_original_scope

    @pyc.register_raw_handler(pyc.argument)
    @pyc.skip_when_tracing_disabled
    def argument(self, arg_obj: Any, arg_node_id: int, *_, **__):
        self.num_args_seen += 1
        arg_node = self.ast_node_by_id.get(arg_node_id, None)
        try:
            mut_cand = self.lexical_call_stack.get_field("mutation_candidate")
        except IndexError:
            return
        if mut_cand is None:
            return

        if (
            isinstance(mut_cand[1], (ListInsert, ListPop, ListRemove))
            and mut_cand[1].pos is None
            and self.num_args_seen == 1
        ):
            try:
                if isinstance(mut_cand[1], ListRemove):
                    mut_obj = mut_cand[0][0]
                    for i in range(len(mut_obj)):
                        if mut_obj[i] == arg_obj:
                            mut_cand[1].pos = i
                            break
                else:
                    mut_cand[1].pos = arg_obj
            except:
                pass

        if isinstance(arg_node, ast.Name):
            assert self.active_scope is self.cur_frame_original_scope
            arg_dsym = self.active_scope.lookup_data_symbol_by_name(arg_node.id)
            if arg_dsym is None:
                self.active_scope.upsert_data_symbol_for_name(
                    arg_node.id,
                    arg_obj,
                    set(),
                    self.prev_trace_stmt_in_cur_frame.stmt_node,
                    implicit=True,
                )
        mut_cand[-2].append(resolve_rval_symbols(arg_node))
        mut_cand[-1].append(arg_obj)

    def _save_mutation_candidate(
        self, obj: Any, method_name: Optional[str], obj_name: Optional[str] = None
    ) -> None:
        mutation_event = resolve_mutating_method(obj, method_name)
        if mutation_event is None or isinstance(
            mutation_event, MutatingMethodEventNotYetImplemented
        ):
            mutation_event = StandardMutation()
        self.mutation_candidate = ((obj, obj_name, method_name), mutation_event, [], [])

    @pyc.register_raw_handler(pyc.before_call)
    @pyc.skip_when_tracing_disabled
    def before_call(self, function_or_method, *_, **__):
        if self.saved_complex_symbol_load_data is None:
            obj, attr_or_subscript, is_subscript, obj_name = None, None, None, None
        else:
            # TODO: this will cause errors if we add more fields
            (
                _,
                obj,
                attr_or_subscript,
                is_subscript,
                *_,
                obj_name,
            ) = self.saved_complex_symbol_load_data
        if obj is not None and is_subscript is not None:
            if is_subscript:
                # TODO: need to do this also for chained calls, e.g. f()()
                method_name = None
            else:
                assert isinstance(attr_or_subscript, str)
                method_name = attr_or_subscript
                # method_name should match ast_by_id[function_or_method].func.id
            self._save_mutation_candidate(obj, method_name, obj_name=obj_name)
        self.saved_complex_symbol_load_data = None
        with self.lexical_call_stack.push():
            pass
        self.active_scope = self.cur_frame_original_scope

    @pyc.register_raw_handler((pyc.before_function_body, pyc.before_lambda_body))
    def before_function_body(self, _obj: Any, function_id: NodeId, *_, **__):
        ret = self.is_tracing_enabled and function_id not in self._seen_functions_ids
        if ret:
            self._seen_functions_ids.add(function_id)
        return ret

    @pyc.register_raw_handler(pyc.after_call)
    def after_call(
        self,
        retval: Any,
        _node_id: NodeId,
        frame: FrameType,
        *_,
        call_node_id: NodeId,
        **__,
    ):
        tracing_will_be_enabled_by_end = self.is_tracing_enabled
        if not self.is_tracing_enabled:
            tracing_will_be_enabled_by_end = self._should_attempt_to_reenable_tracing(
                frame
            )
            if tracing_will_be_enabled_by_end:
                # if tracing gets reenabled here instead of at the 'before_stmt' handler, then we're still
                # at the same module stmt as when tracing was disabled, and we still have a 'return' to trace
                self.call_depth = 1
                self.call_stack.clear()
                self.lexical_call_stack.clear()

        if not tracing_will_be_enabled_by_end:
            return

        # no need to reset active scope here;
        # that will happen in the 'after chain' handler

        if len(self.lexical_call_stack) > 0:
            # skip / give up if tracing was recently reenabled
            self.lexical_call_stack.pop()
        self.prev_node_id_in_cur_frame_lexical = None
        self._process_possible_mutation(retval)

        if not self.is_tracing_enabled:
            self._enable_tracing()

    # Note: we don't trace set literals
    @pyc.register_raw_handler(
        (
            pyc.before_dict_literal,
            pyc.before_list_literal,
            pyc.before_tuple_literal,
        )
    )
    @pyc.skip_when_tracing_disabled
    def before_literal(self, *_, **__):
        parent_scope = self.active_literal_scope or self.cur_frame_original_scope
        with self.lexical_literal_stack.push():
            self.active_literal_scope = Namespace(
                None, Namespace.ANONYMOUS, parent_scope
            )

    @pyc.register_raw_handler(
        (
            pyc.after_dict_literal,
            pyc.after_list_literal,
            pyc.after_tuple_literal,
        )
    )
    @pyc.skip_when_tracing_disabled
    def after_literal(
        self, literal: Union[dict, list, tuple], node_id: NodeId, *_, **__
    ):
        try:
            self.active_literal_scope.update_obj_ref(literal)
            logger.warning("create literal scope %s", self.active_literal_scope)
            starred_idx = -1
            starred_namespace = None
            outer_deps = set()
            for (i, inner_obj), (
                inner_key_node,
                inner_val_node,
            ) in match_container_obj_or_namespace_with_literal_nodes(
                literal, self.ast_node_by_id[node_id]  # type: ignore
            ):
                # TODO: memoize symbol resolution; otherwise this will be quadratic for deeply nested literals
                if isinstance(inner_val_node, ast.Starred):
                    inner_symbols = set()
                    starred_idx += 1
                    if starred_idx == 0:
                        starred_syms = self.resolve_loaded_symbols(inner_val_node)
                        starred_namespace = (
                            nbs().namespaces.get(starred_syms[0].obj_id, None)
                            if starred_syms
                            else None
                        )
                    if starred_namespace is not None:
                        starred_dep = starred_namespace.lookup_data_symbol_by_name_this_indentation(
                            starred_idx, is_subscript=True
                        )
                        inner_symbols.add(starred_dep)
                else:
                    inner_symbols = resolve_rval_symbols(inner_val_node)
                    if inner_key_node is not None:
                        outer_deps.update(resolve_rval_symbols(inner_key_node))
                self.node_id_to_loaded_symbols.pop(id(inner_val_node), None)
                inner_symbols.discard(None)
                if isinstance(
                    i, (int, str)
                ):  # TODO: perform more general check for SupportedIndexType
                    self.active_literal_scope.upsert_data_symbol_for_name(
                        i,
                        inner_obj,
                        inner_symbols,
                        self.prev_trace_stmt_in_cur_frame.stmt_node,
                        is_subscript=True,
                        implicit=True,
                        # this is necessary in case some literal object got reused,
                        # since as of this comment (2021/08/14) we do not clear
                        # GC'd symbols from the symbol graph
                        propagate=False,
                    )
            self.node_id_to_loaded_literal_scope[node_id] = self.active_literal_scope
            parent_scope: Scope = self.active_literal_scope.parent_scope
            assert parent_scope is not None
            literal_sym = parent_scope.upsert_data_symbol_for_name(
                "<literal_sym_%d>" % id(literal),
                literal,
                outer_deps,
                self.prev_trace_stmt_in_cur_frame.stmt_node,
                is_anonymous=True,
                implicit=True,
                propagate=False,
            )
            self.node_id_to_loaded_symbols[node_id].append(literal_sym)
            return literal
        finally:
            self.lexical_literal_stack.pop()

    @pyc.register_raw_handler(pyc.dict_key)
    @pyc.skip_when_tracing_disabled
    def dict_key(self, obj: Any, key_node_id: NodeId, *_, **__):
        self.node_id_to_saved_dict_key[key_node_id] = obj
        return obj

    @pyc.register_raw_handler(pyc.dict_value)
    @pyc.skip_when_tracing_disabled
    def dict_value(
        self,
        obj: Any,
        value_node_id: NodeId,
        *_,
        key_node_id: NodeId,
        dict_node_id: NodeId,
        **__,
    ):
        scope = self.node_id_to_loaded_literal_scope.pop(value_node_id, None)
        if scope is None:
            return obj
        # if we found a pending literal, assert that it's not dict unpacking
        assert key_node_id is not None
        key_obj = self.node_id_to_saved_dict_key.pop(key_node_id, None)
        if isinstance(key_obj, (str, int)):
            scope.scope_name = str(key_obj)
        return obj

    @pyc.register_raw_handler((pyc.list_elt, pyc.tuple_elt))
    @pyc.skip_when_tracing_disabled
    def list_or_tuple_elt(
        self,
        obj: Any,
        elt_node_id: NodeId,
        *_,
        index: Optional[int],
        container_node_id: NodeId,
        **__,
    ):
        scope = self.node_id_to_loaded_literal_scope.pop(elt_node_id, None)
        if scope is None:
            return obj
        if index is not None:
            scope.scope_name = str(index)
        return obj

    @pyc.register_raw_handler(pyc.after_lambda)
    @pyc.skip_when_tracing_disabled
    def after_lambda(self, obj: Any, lambda_node_id: int, frame: FrameType, *_, **__):
        sym_deps = []
        node = self.ast_node_by_id[lambda_node_id]
        for kw_default in node.args.defaults:  # type: ignore
            sym_deps.extend(self.resolve_loaded_symbols(kw_default))
        sym = self.active_scope.upsert_data_symbol_for_name(
            "<lambda_sym_%d>" % id(obj),
            obj,
            sym_deps,
            self.prev_trace_stmt_in_cur_frame.stmt_node,
            is_function_def=True,
            propagate=False,
        )
        # FIXME: this is super brittle. We're passing in a stmt node to update the mapping from
        #  stmt_node to function symbol, but simultaneously forcing the lambda symbol to hold
        #  a reference to the lambda in order to help with symbol resolution later
        sym.stmt_node = node
        self.node_id_to_loaded_symbols[lambda_node_id].append(sym)

    @pyc.register_raw_handler(pyc.after_stmt)
    def after_stmt(self, ret_expr: Any, stmt_id: int, frame: FrameType, *_, **__):
        if stmt_id in self.seen_stmts:
            return ret_expr
        self._saved_stmt_ret_expr = ret_expr
        stmt = self.ast_node_by_id.get(stmt_id, None)
        if stmt is not None:
            self.handle_other_sys_events(
                None, 0, frame, pyc.after_stmt, stmt_node=cast(ast.stmt, stmt)
            )
        return ret_expr

    @pyc.register_raw_handler(pyc.after_module_stmt)
    def after_module_stmt(self, *_, **__):
        if self.is_tracing_enabled:
            assert self.cur_frame_original_scope.is_global
        ret = self._saved_stmt_ret_expr
        self._saved_stmt_ret_expr = None
        self._module_stmt_counter += 1
        return ret

    @pyc.register_raw_handler(pyc.before_stmt)
    def before_stmt(self, _ret: None, stmt_id: int, frame: FrameType, *_, **__) -> None:
        self.next_stmt_node_id = stmt_id
        if stmt_id in self.seen_stmts:
            return
        # logger.warning('reenable tracing: %s', site_id)
        if self.prev_trace_stmt_in_cur_frame is not None:
            prev_trace_stmt_in_cur_frame = self.prev_trace_stmt_in_cur_frame
            # both of the following stmts should be processed when body is entered
            if isinstance(
                prev_trace_stmt_in_cur_frame.stmt_node, (ast.For, ast.If, ast.With)
            ):
                self.after_stmt(None, prev_trace_stmt_in_cur_frame.stmt_id, frame)
        trace_stmt = self.traced_statements.get(stmt_id, None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(
                frame, cast(ast.stmt, self.ast_node_by_id[stmt_id])
            )
            self.traced_statements[stmt_id] = trace_stmt
        self.prev_trace_stmt_in_cur_frame = trace_stmt
        if not self.is_tracing_enabled and self._should_attempt_to_reenable_tracing(
            frame
        ):
            # At this point, we can be sure we're at the top level
            # because tracing was enabled in a top-level handler.
            # We also need to clear the stack, as we won't catch
            # the return event (since tracing was already disabled
            # when we got to a `before_stmt` event).
            self.call_depth = 0
            self.call_stack.clear()
            self.lexical_call_stack.clear()
            self.after_stmt_reset_hook()
            self._enable_tracing()

    def _should_attempt_to_reenable_tracing(self, frame: FrameType) -> bool:
        if nbs().is_develop:
            assert not self.is_tracing_enabled
            assert self.call_depth > 0, (
                "expected managed call depth > 0, got %d" % self.call_depth
            )
        call_depth = 0
        while frame is not None:
            if nbs().is_cell_file(frame.f_code.co_filename):
                call_depth += 1
            frame = frame.f_back
        if nbs().is_develop:
            assert call_depth >= 1, "expected call depth >= 1, got %d" % call_depth
        # TODO: allow reenabling tracing beyond just at the top level
        if call_depth != 1:
            return False
        if len(self.call_stack) == 0:
            stmt_in_top_level_frame = self.prev_trace_stmt_in_cur_frame
        else:
            stmt_in_top_level_frame = self.call_stack.get_field(
                "prev_trace_stmt_in_cur_frame", depth=0
            )
        if stmt_in_top_level_frame.finished:
            return False
        if nbs().trace_messages_enabled:
            self.EVENT_LOGGER.warning("reenable tracing >>>")
        return True

    def _get_or_make_trace_stmt(
        self, stmt_node: ast.stmt, frame: FrameType
    ) -> TraceStatement:
        trace_stmt = self.traced_statements.get(id(stmt_node), None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(frame, stmt_node)
            self.traced_statements[id(stmt_node)] = trace_stmt
        return trace_stmt

    def _maybe_log_event(
        self, event: pyc.TraceEvent, stmt_node: ast.stmt, trace_stmt: TraceStatement
    ):
        if nbs().trace_messages_enabled:
            codeline = astunparse.unparse(stmt_node).strip("\n").split("\n")[0]
            codeline = " " * getattr(stmt_node, "col_offset", 0) + codeline
            self.EVENT_LOGGER.warning(
                " %3d: %10s >>> %s", trace_stmt.lineno, event, codeline
            )

    def _get_stmt_node_for_sys_event(
        self, event: pyc.TraceEvent, cell_num: int, lineno: int
    ) -> Optional[ast.stmt]:
        if event == pyc.return_ and self.next_stmt_node_id is not None:
            # this branch necessary for python < 3.8 where the frame
            # position maps to the calling location instead of the return
            return cast(ast.stmt, self.ast_node_by_id[self.next_stmt_node_id])
        try:
            stmt_node = self.line_to_stmt_by_module_id[cell_num][lineno]
            if event == pyc.call and not isinstance(
                stmt_node, (ast.AsyncFunctionDef, ast.FunctionDef)
            ):
                # TODO: this is bad and I should feel bad. Need a better way to figure out which
                #  stmt is executing than by using line numbers.
                parent_node = self.parent_node_by_id.get(id(stmt_node), None)
                if nbs().is_develop:
                    logger.info(
                        "node %s parent %s",
                        ast.dump(stmt_node),
                        None if parent_node is None else ast.dump(parent_node),
                    )
                if (
                    parent_node is not None
                    and getattr(parent_node, "lineno", None) == lineno
                    and isinstance(parent_node, (ast.AsyncFunctionDef, ast.FunctionDef))
                ):
                    stmt_node = parent_node
            return stmt_node
        except KeyError as e:
            if nbs().is_develop:
                self.EVENT_LOGGER.warning(
                    "got key error for stmt node in cell %d, line %d",
                    cell_num,
                    lineno,
                )
                raise e
        return None

    @pyc.register_raw_handler((pyc.call, pyc.return_))
    def handle_first_ipython_frame(
        self,
        _ret: Any,
        _node_id: None,
        _frame: FrameType,
        event: pyc.TraceEvent,
        *_,
        **__,
    ):
        # IPython quirk -- every line in outer scope apparently wrapped in lambda
        # We want to skip the outer 'call' and 'return' for these
        if event == pyc.call:
            self.call_depth += 1
            if self.call_depth == 1:
                return pyc.Skip
        elif event == pyc.return_:
            self.call_depth -= 1
            if nbs().is_develop:
                assert self.call_depth >= 0
            if self.call_depth == 0:
                return pyc.Skip

    @pyc.register_raw_handler(pyc.call)
    def handle_call(
        self,
        ret_obj: Any,
        _node_id: None,
        frame: FrameType,
        event: pyc.TraceEvent,
        *_,
        **__,
    ):
        cell_num, lineno = nbs().get_position(frame)
        assert cell_num is not None
        stmt_node = self._get_stmt_node_for_sys_event(event, cell_num, lineno)
        trace_stmt = self._get_or_make_trace_stmt(stmt_node, frame)
        self._maybe_log_event(event, stmt_node, trace_stmt)

        try:
            prev_node_id_in_cur_frame_lexical = self.lexical_call_stack.get_field(
                "prev_node_id_in_cur_frame_lexical"
            )
        except IndexError:
            # this could happen if the call happens in library code,
            # and the corresponding notebook statement isn't an ast.Call
            # (e.g., it's a property or just induces a __repr__ call)
            # Make node_id_for_last_call point to self to cover such cases
            prev_node_id_in_cur_frame_lexical = id(stmt_node)

        if trace_stmt.node_id_for_last_call == prev_node_id_in_cur_frame_lexical:
            if nbs().trace_messages_enabled:
                self.EVENT_LOGGER.warning(" disable tracing >>>")
            self._disable_tracing()
            return pyc.Null
        trace_stmt.node_id_for_last_call = prev_node_id_in_cur_frame_lexical
        self.state_transition_hook(event, trace_stmt, ret_obj)

    @pyc.register_raw_handler((pyc.return_, pyc.exception))
    def handle_other_sys_events(
        self,
        ret_obj: Any,
        _node_id: None,
        frame: FrameType,
        event: pyc.TraceEvent,
        *_,
        stmt_node: Optional[ast.stmt] = None,
        **__,
    ):
        assert self.is_tracing_enabled or event == pyc.after_stmt

        cell_num, lineno = nbs().get_position(frame)
        assert cell_num is not None

        if event == pyc.after_stmt:
            assert stmt_node is not None
        else:
            stmt_node = self._get_stmt_node_for_sys_event(event, cell_num, lineno)

        trace_stmt = self._get_or_make_trace_stmt(stmt_node, frame)
        self._maybe_log_event(event, stmt_node, trace_stmt)
        self.state_transition_hook(event, trace_stmt, ret_obj)



# -*- coding: utf-8 -*-
import logging
from typing import Optional, Set

from nbsafety.data_model.data_symbol import DataSymbol
from nbsafety.singletons import nbs
from test.utils import assert_bool, make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
_safety_fixture, run_cell_ = make_safety_fixture(
    setup_cells=["from test.test_nested_symbols import DotDict"]
)
run_cell = run_cell_


def stale_detected():
    return nbs().test_and_clear_detected_flag()


def assert_detected(msg=""):
    assert_bool(stale_detected(), msg=msg)


def assert_not_detected(msg=""):
    assert_bool(not stale_detected(), msg=msg)


def lookup_symbols(val) -> Optional[Set[DataSymbol]]:
    safety = nbs()
    alias_set = {
        alias for alias in safety.aliases.get(id(val), []) if not alias.is_anonymous
    }
    if len(alias_set) == 0:
        return None
    return alias_set


def lookup_symbol(val) -> Optional[DataSymbol]:
    alias_set = lookup_symbols(val)
    if alias_set is None or len(alias_set) == 0:
        return None
    alias = None
    for alias in alias_set:
        # try to find one that isn't anonymous or
        # associated with anonymous namespace to avoid
        # e.g. <literal_sym_12345> in tests
        containing_ns = alias.containing_namespace
        if containing_ns is not None and containing_ns.is_anonymous:
            continue
        if alias.is_anonymous:
            continue
        return alias
    return alias


# ref: https://gist.github.com/golobor/397b5099d42da476a4e6
class DotDict(dict):
    """A dict with dot access and autocompletion.

    The idea and most of the code was taken from
    http://stackoverflow.com/a/23689767,
    http://code.activestate.com/recipes/52308-the-simple-but-handy-collector-of-a-bunch-of-named/
    http://stackoverflow.com/questions/2390827/how-to-properly-subclass-dict-and-override-get-set
    """

    def __init__(self, *a, **kw):
        dict.__init__(self)
        self.update(*a, **kw)
        self.__dict__ = self

    def __setattr__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError("This key is reserved for the dict methods.")
        dict.__setattr__(self, key, value)

    def __setitem__(self, key, value):
        if key in dict.__dict__:
            raise AttributeError("This key is reserved for the dict methods.")
        dict.__setitem__(self, key, value)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v

    def __getstate__(self):
        return self

    def __setstate__(self, state):
        self.update(state)
        self.__dict__ = self


def test_basic():
    run_cell("d = DotDict()")
    run_cell("d.x = DotDict()")
    run_cell("d.y = DotDict()")
    run_cell("d.x.a = 5")
    # run_cell('d.x.b = 6')
    run_cell("logging.info(d.x.a)")
    assert_not_detected()
    run_cell("logging.info(d.x)")
    assert_not_detected()
    run_cell("d.y.a = 7")
    # run_cell('d.y.b = 8')
    run_cell("logging.info(d.y.a)")
    assert_not_detected()
    run_cell("logging.info(d.y)")
    assert_not_detected()
    run_cell("x = d.x.a + 9")
    run_cell("y = d.y.a + 9")
    run_cell("d.y.a = 9")
    run_cell("logging.info(x)")
    assert_not_detected("`x` independent of changed `d.y.a`")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on changed `d.y.a`")
    run_cell("logging.info(d.x.a)")
    assert_not_detected()
    run_cell("logging.info(d.x)")
    assert_not_detected()
    run_cell("logging.info(d)")
    assert_not_detected()
    run_cell("d.y = 10")
    run_cell("logging.info(x)")
    assert_not_detected("`x` independent of changed `d.y`")
    run_cell("logging.info(y)")
    assert_detected("`y` depends on changed `d.y`")


def test_nested_readable_name():
    run_cell("d = DotDict()")
    run_cell("d.x = DotDict()")
    run_cell("d.x.a = 5")
    run_cell("d.x.b = 6")
    d_x_a = lookup_symbol(5)
    assert d_x_a.readable_name == "d.x.a"
    d_x_b = lookup_symbol(6)
    assert d_x_b.readable_name == "d.x.b"
    run_cell("d.y = DotDict()")
    run_cell("d.y.a = 7")
    run_cell("d.y.b = 8")
    d_y_a = lookup_symbol(7)
    assert d_y_a.readable_name == "d.y.a"
    d_y_b = lookup_symbol(8)
    assert d_y_b.readable_name == "d.y.b"


def test_nested_readable_name_dict_literal():
    run_cell('d = {"x": {"y": 42}}')
    d_x_y = lookup_symbol(42)
    assert d_x_y.readable_name == "d[x][y]", (
        "got %s when expected d[x][y]" % d_x_y.readable_name
    )


def test_nested_readable_name_list_literal():
    run_cell("lst = [0, [1, 2, 3]]")
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == "lst[1][1]", (
        "got %s when expected lst[1][1]" % lst_1_1.readable_name
    )


def test_nested_readable_name_tuple_in_list():
    run_cell("lst = [0, (1, 2, 3)]")
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == "lst[1][1]", (
        "got %s when expected lst[1][1]" % lst_1_1.readable_name
    )


def test_nested_readable_name_list_in_tuple():
    run_cell("lst = (0, [1, 2, 3])")
    lst_1_1 = lookup_symbol(2)
    assert lst_1_1.readable_name == "lst[1][1]", (
        "got %s when expected lst[1][1]" % lst_1_1.readable_name
    )


def test_nested_symbol_created_for_symbol_already_existing():
    run_cell("x = 42")
    run_cell("lst = [1, 2, [7, x, 8], 4]")
    x_syms = lookup_symbols(42)
    assert len(x_syms) == 2
    x_names = {sym.readable_name for sym in x_syms}
    assert "x" in x_names, "could not find `x` in set of names %s" % x_names
    assert "lst[2][1]" in x_names, (
        "could not find `lst[2][1]` in set of names %s" % x_names
    )


def test_list_append_extend():
    run_cell("lst = []")
    run_cell("lst.append(42)")
    run_cell("lst.extend([43, 44])")
    name = lookup_symbol(42).readable_name
    assert name == "lst[0]", "got %s" % name
    name = lookup_symbol(43).readable_name
    assert name == "lst[1]", "got %s" % name
    name = lookup_symbol(44).readable_name
    assert name == "lst[2]", "got %s" % name


def test_list_insert():
    run_cell("lst = [0, 1, 2, 4, 5, 6]")
    name = lookup_symbol(2).readable_name
    assert name == "lst[2]", "got %s" % name
    name = lookup_symbol(4).readable_name
    assert name == "lst[3]", "got %s" % name
    dsym = lookup_symbol(3)
    assert dsym is None
    run_cell("lst.insert(3, 3)")
    sym_2 = lookup_symbol(2)
    assert sym_2.readable_name == "lst[2]", "got %s" % sym_2.readable_name
    assert sym_2.obj == 2
    sym_3 = lookup_symbol(3)
    assert sym_3.readable_name == "lst[3]", "got %s" % sym_3.readable_name
    assert sym_3.obj == 3
    sym_4 = lookup_symbol(4)
    assert sym_4.readable_name == "lst[4]", "got %s" % sym_4.readable_name
    assert sym_4.obj == 4
    assert (
        sym_4.containing_namespace.lookup_data_symbol_by_name_this_indentation(
            4, is_subscript=True
        )
        is sym_4
    )


def test_list_delete():
    run_cell("lst = [0, 1, 2, 3, 3, 4, 5, 6]")
    name = lookup_symbol(2).readable_name
    assert name == "lst[2]", "got %s" % name
    name = lookup_symbol(4).readable_name
    assert name == "lst[5]", "got %s" % name
    run_cell("del lst[3]")
    sym_2 = lookup_symbol(2)
    assert sym_2.readable_name == "lst[2]", "got %s" % sym_2.readable_name
    sym_3 = lookup_symbol(3)
    assert sym_3.readable_name == "lst[3]", "got %s" % sym_3.readable_name
    sym_4 = lookup_symbol(4)
    assert sym_4.readable_name == "lst[4]", "got %s" % sym_4.readable_name
    assert (
        sym_4.containing_namespace.lookup_data_symbol_by_name_this_indentation(
            4, is_subscript=True
        )
        is sym_4
    )



# -*- coding: utf-8 -*-
import difflib
import functools
import logging
from types import FrameType
from typing import List, Set, Union

import hypothesis.strategies as st
from hypothesis import example, given, settings

from nbsafety.safety import NotebookSafety
from nbsafety.singletons import tracer
from nbsafety.tracing.nbsafety_tracer import SafetyTracer
from pyccolo import TraceEvent
from .utils import make_safety_fixture, skipif_known_failing


logging.basicConfig(level=logging.INFO)


NotebookSafety.instance()
_ALL_EVENTS_WITH_HANDLERS = SafetyTracer.instance().events_with_registered_handlers
_RECORDED_EVENTS = []


def subsets(draw, elements):
    return {e for e in elements if draw(st.booleans())}


def patched_emit_event_fixture():
    _RECORDED_EVENTS.clear()
    original_emit_event = SafetyTracer._emit_event

    def _patched_emit_event(
        self, evt: Union[str, TraceEvent], node_id: int, frame: FrameType, **kwargs
    ):
        event = evt if isinstance(evt, TraceEvent) else TraceEvent(evt)
        if frame.f_code.co_filename.startswith("<ipython-input"):
            if not (
                (event == TraceEvent.call and self.call_depth == 0)
                or (event == TraceEvent.return_ and self.call_depth == 1)
            ):
                if event in self.events_with_registered_handlers:
                    _RECORDED_EVENTS.append(event)
        return original_emit_event(self, event, node_id, frame, **kwargs)

    SafetyTracer._emit_event = _patched_emit_event
    yield
    SafetyTracer._emit_event = original_emit_event


# Reset dependency graph before each test
_safety_fixture, run_cell_ = make_safety_fixture(
    extra_fixture=patched_emit_event_fixture,
    # trace_messages_enabled=True,
)


_DIFFER = difflib.Differ()


def patch_events_with_registered_handlers_to_subset(testfunc):
    @functools.wraps(testfunc)
    @settings(max_examples=20, deadline=None)
    @example(events=set(_ALL_EVENTS_WITH_HANDLERS))
    def wrapped_testfunc(events):
        events |= {
            TraceEvent.before_subscript_load,
            TraceEvent.after_subscript_load,
            TraceEvent.before_subscript_store,
            TraceEvent.before_subscript_del,
            TraceEvent._load_saved_slice,
            TraceEvent.before_load_complex_symbol,
            TraceEvent.after_load_complex_symbol,
            TraceEvent.before_attribute_load,
            TraceEvent.after_attribute_load,
            TraceEvent.before_attribute_store,
            TraceEvent.before_attribute_del,
            TraceEvent.before_call,
            TraceEvent.after_call,
            TraceEvent.argument,
            TraceEvent.before_return,
            TraceEvent.after_return,
            TraceEvent.call,
            TraceEvent.return_,
            TraceEvent.exception,
            TraceEvent.before_stmt,
            TraceEvent.after_stmt,
            TraceEvent.after_assign_rhs,
        }
        list_literal_related = {
            TraceEvent.before_list_literal,
            TraceEvent.after_list_literal,
            TraceEvent.list_elt,
        }
        if events & list_literal_related:
            events |= list_literal_related
        set_literal_related = {
            TraceEvent.before_set_literal,
            TraceEvent.after_set_literal,
            TraceEvent.set_elt,
        }
        if events & set_literal_related:
            events |= set_literal_related
        tuple_literal_related = {
            TraceEvent.before_tuple_literal,
            TraceEvent.after_tuple_literal,
            TraceEvent.tuple_elt,
        }
        if events & tuple_literal_related:
            events |= tuple_literal_related
        dict_literal_related = {
            TraceEvent.before_dict_literal,
            TraceEvent.after_dict_literal,
            TraceEvent.dict_key,
            TraceEvent.dict_value,
        }
        if events & dict_literal_related:
            events |= dict_literal_related

        orig_handlers = tracer().events_with_registered_handlers
        try:
            tracer().events_with_registered_handlers = frozenset(events)
            _RECORDED_EVENTS.clear()
            testfunc(events)
        finally:
            tracer().events_with_registered_handlers = orig_handlers

    return wrapped_testfunc


def filter_events_to_subset(
    events: List[TraceEvent], subset: Set[TraceEvent]
) -> List[TraceEvent]:
    return [evt for evt in events if evt in subset]


def throw_and_print_diff_if_recorded_not_equal_to(actual: List[TraceEvent]) -> None:
    assert _RECORDED_EVENTS == actual, "\n".join(
        _DIFFER.compare(
            [evt.value for evt in _RECORDED_EVENTS], [evt.value for evt in actual]
        )
    )
    _RECORDED_EVENTS.clear()


def run_cell(cell, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, **kwargs)


@st.composite
def subsets(draw, elements):
    return {e for e in elements if draw(st.booleans())} | {
        TraceEvent.init_module,
        TraceEvent.before_stmt,
        TraceEvent.after_stmt,
        TraceEvent.after_module_stmt,
        TraceEvent.call,
        TraceEvent.return_,
        TraceEvent.after_for_loop_iter,
        TraceEvent.after_while_loop_iter,
    }


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_recorded_events_simple(events):
    assert _RECORDED_EVENTS == []
    run_cell('logging.info("foo")')
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_recorded_events_two_stmts(events):
    assert _RECORDED_EVENTS == []
    run_cell("x = [1, 2, 3]")
    run_cell("logging.info(x)")
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.before_list_literal,
                *([TraceEvent.list_elt] * 3),
                TraceEvent.after_list_literal,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.load_name,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_nested_chains_no_call(events):
    assert _RECORDED_EVENTS == []
    run_cell('logging.info("foo is %s", logging.info("foo"))')
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.argument,
                # next events correspond to `logging.info("foo")`
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_list_nested_in_dict(events):
    assert _RECORDED_EVENTS == []
    run_cell("x = {1: [2, 3, 4]}")
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.before_dict_literal,
                TraceEvent.dict_key,
                TraceEvent.before_list_literal,
                *([TraceEvent.list_elt] * 3),
                TraceEvent.after_list_literal,
                TraceEvent.dict_value,
                TraceEvent.after_dict_literal,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_function_call(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        def foo(x):
            return [x]
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )
    run_cell("foo([42])")
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.before_list_literal,
                TraceEvent.list_elt,
                TraceEvent.after_list_literal,
                TraceEvent.argument,
                TraceEvent.call,
                TraceEvent.before_function_body,
                TraceEvent.before_stmt,
                TraceEvent.before_return,
                TraceEvent.before_list_literal,
                TraceEvent.load_name,
                TraceEvent.list_elt,
                TraceEvent.after_list_literal,
                TraceEvent.after_return,
                TraceEvent.after_function_execution,
                TraceEvent.return_,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_lambda_in_tuple(events):
    assert _RECORDED_EVENTS == []
    run_cell("x = (lambda: 42,)")
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.before_tuple_literal,
                TraceEvent.before_lambda,
                TraceEvent.after_lambda,
                TraceEvent.tuple_elt,
                TraceEvent.after_tuple_literal,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_fancy_slices(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        import numpy as np
        class Foo:
            def __init__(self, x):
                self.x = x
        foo = Foo(1)
        arr = np.zeros((3, 3, 3))
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                # import numpy as np
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # class Foo: ...
                TraceEvent.before_stmt,
                TraceEvent.call,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.return_,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # foo = Foo(1)
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.call,
                TraceEvent.before_function_body,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.load_name,
                TraceEvent.after_assign_rhs,
                TraceEvent.load_name,
                TraceEvent.before_attribute_store,
                TraceEvent.after_stmt,
                TraceEvent.after_function_execution,
                TraceEvent.return_,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # arr = np.zeros((3, 3, 3))
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.before_tuple_literal,
                TraceEvent.tuple_elt,
                TraceEvent.tuple_elt,
                TraceEvent.tuple_elt,
                TraceEvent.after_tuple_literal,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )

    run_cell("logging.info(arr[foo.x:foo.x+1,...])")
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.before_call,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_attribute_load,
                TraceEvent.after_attribute_load,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_subscript_slice,
                TraceEvent.before_subscript_load,
                TraceEvent._load_saved_slice,
                TraceEvent.after_subscript_load,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_for_loop(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        for i in range(10):
            pass
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
            ]
            + [
                TraceEvent.before_for_loop_body,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_for_loop_iter,
                # ] * 10 + [
            ]
            * 1
            + [
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_while_loop(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        i = 0
        while i < 10:
            i += 1
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                TraceEvent.before_stmt,
            ]
            + [
                TraceEvent.load_name,
                TraceEvent.before_while_loop_body,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_while_loop_iter,
                # ] * 10 + [
            ]
            * 1
            + [
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_loop_with_continue(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        for i in range(10):
            continue
            print("hi")
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.before_for_loop_body,
                TraceEvent.before_stmt,
                TraceEvent.after_for_loop_iter,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_for_loop_nested_in_while_loop(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        i = 0
        while i < 10:
            for j in range(2):
                i += 1
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                TraceEvent.before_stmt,
                TraceEvent.before_assign_rhs,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                TraceEvent.before_stmt,
            ]
            + [
                TraceEvent.load_name,
                TraceEvent.before_while_loop_body,
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.argument,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.before_for_loop_body,
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_for_loop_iter,
                # TraceEvent.before_stmt,
                # TraceEvent.after_stmt,
                # TraceEvent.after_loop_iter,
                TraceEvent.after_stmt,
                TraceEvent.after_while_loop_iter,
                # ] * 5 + [
            ]
            * 1
            + [
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )


@given(events=subsets(_ALL_EVENTS_WITH_HANDLERS))
@patch_events_with_registered_handlers_to_subset
def test_lambda_wrapping_call(events):
    assert _RECORDED_EVENTS == []
    run_cell(
        """
        z = 42
        def f():
            return z
        lam = lambda: f()
        x = lam()
        """
    )
    throw_and_print_diff_if_recorded_not_equal_to(
        filter_events_to_subset(
            [
                TraceEvent.init_module,
                # z = 42
                TraceEvent.before_stmt,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # def f(): ...
                TraceEvent.before_stmt,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # lam = lambda: f()
                TraceEvent.before_stmt,
                TraceEvent.before_lambda,
                TraceEvent.after_lambda,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
                # x = lam()
                TraceEvent.before_stmt,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.call,
                TraceEvent.before_lambda_body,
                TraceEvent.before_load_complex_symbol,
                TraceEvent.load_name,
                TraceEvent.before_call,
                TraceEvent.call,
                TraceEvent.before_function_body,
                TraceEvent.before_stmt,
                TraceEvent.before_return,
                TraceEvent.load_name,
                TraceEvent.after_return,
                TraceEvent.after_function_execution,
                TraceEvent.return_,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.return_,
                TraceEvent.after_call,
                TraceEvent.after_load_complex_symbol,
                TraceEvent.after_assign_rhs,
                TraceEvent.after_stmt,
                TraceEvent.after_module_stmt,
            ],
            events,
        )
    )



# -*- coding: utf-8 -*-
import ast
from contextlib import contextmanager
from typing import Callable, List, Optional

from IPython import get_ipython


def _ipython():
    return get_ipython()


class _IpythonState:
    def __init__(self):
        self.cell_counter: Optional[int] = None

    @contextmanager
    def save_number_of_currently_executing_cell(self):
        self.cell_counter = _ipython().execution_count
        yield
        self.cell_counter = None

    @contextmanager
    def ast_transformer_context(self, transformers: List[ast.NodeTransformer]):
        old = _ipython().ast_transformers
        _ipython().ast_transformers = old + transformers
        yield
        _ipython().ast_transformers = old

    @contextmanager
    def input_transformer_context(
        self, transformers: List[Callable[[List[str]], List[str]]]
    ):
        old = _ipython().input_transformers_post
        _ipython().input_transformers_post = old + transformers
        yield
        _ipython().input_transformers_post = old


_IPY = _IpythonState()


def save_number_of_currently_executing_cell():
    return _IPY.save_number_of_currently_executing_cell()


def ast_transformer_context(transformers):
    return _IPY.ast_transformer_context(transformers)


def input_transformer_context(transformers):
    return _IPY.input_transformer_context(transformers)


def cell_counter() -> int:
    if _IPY.cell_counter is None:
        raise ValueError("should be inside context manager here")
    return _IPY.cell_counter


def run_cell(cell, **kwargs):
    return _ipython().run_cell(
        cell,
        store_history=kwargs.pop("store_history", True),
        silent=kwargs.pop("silent", False),
    )



# -*- coding: utf-8 -*-
import logging
from contextlib import contextmanager
from typing import Dict

from nbsafety.safety import NotebookSafetySettings
from nbsafety.singletons import nbs
from test.utils import make_safety_fixture, skipif_known_failing

logging.basicConfig(level=logging.ERROR)

# Reset dependency graph before each test
# _safety_fixture, run_cell_ = make_safety_fixture(trace_messages_enabled=True)
_safety_fixture, run_cell_ = make_safety_fixture()


@contextmanager
def override_settings(**kwargs):
    old_settings = nbs().settings
    new_settings = old_settings._asdict()
    new_settings.update(kwargs)
    new_settings = NotebookSafetySettings(**new_settings)
    try:
        nbs().settings = new_settings
        yield
    finally:
        nbs().settings = old_settings


def run_cell(cell, cell_id, **kwargs):
    # print()
    # print('*******************************************')
    # print('running', cell)
    # print('*******************************************')
    # print()
    run_cell_(cell, cell_id=cell_id, **kwargs)


def run_all_cells(cells: Dict[int, str], **kwargs):
    for cell_id in sorted(cells.keys()):
        run_cell(cells[cell_id], cell_id=cell_id, **kwargs)


def test_simple():
    cells = {
        0: "x = 0",
        1: "y = x + 1",
        2: "logging.info(y)",
        3: "x = 42",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    assert response.stale_links == {2: {1}}
    assert response.refresher_links == {1: {2}}


def test_refresh_after_exception_fixed():
    cells = {
        0: "x = 0",
        1: "y = x + 1",
        2: "logging.info(y)",
    }
    run_cell(cells[0], 0)
    run_cell(cells[2], 2, ignore_exceptions=True)
    run_cell(cells[1], 1)
    response = nbs().check_and_link_multiple_cells()
    assert response.fresh_cells == {2}


def test_refresh_after_val_changed():
    cells = {
        0: "x = 0",
        1: "y = x + 1",
        2: "logging.info(y)",
        3: "y = 42",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.fresh_cells == {2}


def test_inner_mutation_considered_fresh():
    cells = {
        0: "lst_0 = [0,1,2]",
        1: "lst_1 = [3,4,5]",
        2: "lst = [lst_0, lst_1]",
        3: "logging.info(lst)",
        4: "lst_0.append(42)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


# @pytest.mark.parametrize("force_subscript_symbol_creation", [True, False])
def test_update_list_elem():
    force_subscript_symbol_creation = True
    cells = {
        0: (
            """
            class Foo:
                def __init__(self):
                    self.counter = 0
                    self.dummy = 0
                    
                def inc(self):
                    self.counter += 1
            """
        ),
        1: (
            """
            lst = []
            for i in range(5):
                x = Foo()
                lst.append(x)
            """
        ),
        2: (
            """
            for foo in lst:
                foo.inc()
            """
        ),
        3: "logging.info(lst)",
    }

    run_all_cells(cells)

    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), "got %s" % response.fresh_cells

    cells[4] = "x.inc()"
    run_cell(cells[4], 4)

    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set(), "got %s" % response.stale_cells
    assert response.fresh_cells == {2, 3}, "got %s" % response.fresh_cells

    cells[5] = "foo.inc()"
    run_cell(cells[5], 5)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set(), "got %s" % response.stale_cells
    assert response.fresh_cells == {2, 3, 4}, "got %s" % response.fresh_cells

    if force_subscript_symbol_creation:
        cells[6] = "lst[-1]"
        run_cell(cells[6], 6)
        response = nbs().check_and_link_multiple_cells()
        assert response.stale_cells == set()
        assert response.fresh_cells == {2, 3, 4}

    run_cell(cells[4], 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(
        [2, 3, 5] + ([6] if force_subscript_symbol_creation else [])
    )


def test_no_freshness_for_alias_assignment_post_mutation():
    cells = {
        0: "x = []",
        1: "y = x",
        2: "x.append(5)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()


def test_fresh_after_import():
    cells = {0: "x = np.random.random(10)", 1: "import numpy as np"}
    run_all_cells(cells, ignore_exceptions=True)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {0}


def test_external_object_update_propagates_to_stale_namespace_symbols():
    cells = {
        0: "import fakelib",
        1: "foo = fakelib.Foo()",
        2: "logging.info(foo.x)",
        3: "x = 42",
        4: "foo.x = x + 1",
        5: "x = 43",
        6: "foo = foo.set_x(10)",
    }
    with override_settings(mark_stale_symbol_usages_unsafe=False):
        run_all_cells(cells)
        response = nbs().check_and_link_multiple_cells()
        assert response.stale_cells == set(), "got %s" % response.stale_cells
        assert response.fresh_cells == {2, 4}


def test_symbol_on_both_sides_of_assignment():
    cells = {
        0: "x = 0",
        1: "y = x + 1",
        2: "y += 7",
        3: "x = 42",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    assert list(response.refresher_links.keys()) == [1]


def test_updated_namespace_after_subscript_dep_removed():
    cells = {
        0: "x = 5",
        1: "d = {x: 5}",
        2: "logging.info(d[5])",
        3: "x = 9",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}
    cells[1] = "d = {5: 6}"
    run_cell(cells[1], 1)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell(cells[2], 2)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell(cells[0], 0)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), "got %s" % response.fresh_cells


def test_equal_list_update_does_induce_fresh_cell():
    cells = {
        0: 'x = ["f"] + ["o"] * 10',
        1: 'y = x + list("bar")',
        2: "logging.info(y)",
        3: 'y = list("".join(y))',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell('y = ("f",)', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


def test_equal_list_update_does_induce_fresh_cell_LITERAL_WITH_F_IS_REUSED_ON_UBUNTU_20_04_2_PYTHON_3_8_11():
    cells = {
        0: 'x = ["f"] + ["o"] * 10',
        1: 'y = x + list("bar")',
        2: "logging.info(y)",
        3: 'y = list("".join(y))',
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}
    run_cell('y = ["f"]', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set(), "got %s" % response.stale_cells
    assert response.fresh_cells == {2, 3}


def test_equal_dict_update_does_induce_fresh_cell():
    cells = {
        0: 'x = {"foo": 42, "bar": 43}',
        1: 'y = dict(set(x.items()) | set({"baz": 44}.items()))',
        2: "logging.info(y)",
        3: "y = dict(y.items())",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}, "got %s" % response.fresh_cells
    run_cell('y = {"foo": 99}', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2, 3}


def test_list_append():
    cells = {
        0: "lst = [0, 1]",
        1: "x = lst[1] + 1",
        2: "logging.info(x)",
        3: "lst.append(2)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("lst[1] += 42", 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_list_extend():
    cells = {
        0: "lst = [0, 1]",
        1: "x = lst[1] + 1",
        2: "logging.info(x)",
        3: "lst.extend([2, 3, 4])",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("lst[1] += 42", 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_implicit_subscript_symbol_does_not_bump_ts():
    cells = {
        0: "lst = [] + [0, 1]",
        1: "logging.info(lst)",
        2: "logging.info(lst[0])",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()


def test_liveness_skipped_for_simple_assignment_involving_aliases():
    cells = {
        0: "lst = [1, 2, 3]",
        1: "lst2 = lst",
        2: "lst.append(4)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("lst = [1, 2, 3, 4]", 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {1, 2}, "got %s" % response.fresh_cells


def test_incorrect_object_not_used_for_argument_symbols():
    cells = {
        0: "import numpy as np",
        1: "arr = np.random.randn(100)",
        2: "def f(x): return []",
        # when obj for `np` was passed to `x`, this created issues
        3: "f(np.arange(10))",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set(), "got %s" % response.fresh_cells


def test_increment_by_same_amount():
    cells = {
        0: "x = 2",
        1: "y = x + 1",
        2: "logging.info(y)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("x = 3", 0)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {2}
    assert response.fresh_cells == {1}


def test_list_insert():
    cells = {
        0: "lst = [0, 1, 2, 4, 5, 6]",
        1: "logging.info(lst[0])",
        2: "logging.info(lst[1])",
        3: "logging.info(lst[2])",
        4: "logging.info(lst[3])",
        5: "logging.info(lst[4])",
        6: "logging.info(lst[5])",
        7: "x = lst[5] + 42",
        8: "logging.info(x)",
        9: "lst.insert(3, 3)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {8}, "got %s" % response.stale_cells
    assert response.fresh_cells == {4, 5, 6, 7}, "got %s" % response.fresh_cells


def _test_list_delete_helper(last_cell):
    cells = {
        0: "lst = [0, 1, 2, 3, 3, 4, 5, 6]",
        1: "logging.info(lst[0])",
        2: "logging.info(lst[1])",
        3: "logging.info(lst[2])",
        4: "logging.info(lst[3])",
        5: "logging.info(lst[4])",
        6: "logging.info(lst[5])",
        7: "logging.info(lst[6])",
        8: "logging.info(lst[7])",
        9: "x = lst[6] + 42",
        10: "logging.info(x)",
        11: last_cell,
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == {10}
    # TODO: ideally we would detect that lst[3] is the same after deleting
    #  and not consider cell 4 to be fresh
    # TODO: ideally cell 8 would be considered unsafe since the last entry
    #  no longer exists
    assert response.fresh_cells == {4, 5, 6, 7, 9}, "got %s" % response.fresh_cells


def test_list_delete():
    _test_list_delete_helper("del lst[3]")


def test_list_pop():
    _test_list_delete_helper("lst.pop(3)")


def test_list_remove():
    _test_list_delete_helper("lst.remove(3)")


def test_list_clear():
    cells = {
        0: "lst = [0]",
        1: "logging.info(lst[0])",
        2: "logging.info(lst)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("lst.clear()", 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}


def test_dict_clear():
    cells = {
        0: "d = {0: 0}",
        1: "logging.info(d[0])",
        2: "logging.info(d)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell("d.clear()", 3)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {2}


def test_adhoc_pandas_series_update():
    cells = {
        0: "import pandas as pd",
        1: "df = pd.DataFrame({1: [2, 3], 4: [5, 7]})",
        2: 'df["foo"] = [8, 9]',
        3: "df.foo.dropna(inplace=True)",
    }
    run_all_cells(cells)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == set()
    run_cell('df["foo"] = [8, 9]', 4)
    response = nbs().check_and_link_multiple_cells()
    assert response.stale_cells == set()
    assert response.fresh_cells == {3}



# -*- coding: utf-8 -*-
import ast
import logging
from typing import (
    cast,
    TYPE_CHECKING,
    Any,
    Generator,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from nbsafety.analysis.resolved_symbols import ResolvedDataSymbol
from nbsafety.singletons import tracer
from nbsafety.utils.ast_utils import subscript_to_slice
from nbsafety.utils import CommonEqualityMixin
from nbsafety.types import SupportedIndexType

if TYPE_CHECKING:
    from nbsafety.data_model.data_symbol import DataSymbol, Scope


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def resolve_slice_to_constant(
    node: ast.Subscript,
) -> Optional[Union[SupportedIndexType, ast.Name]]:
    """
    Version-independent way to get at the slice data
    """
    slc = subscript_to_slice(node)

    if isinstance(slc, ast.Tuple):
        elts: Any = []
        for v in slc.elts:
            if isinstance(v, ast.Num):
                elts.append(v.n)
            elif isinstance(v, ast.Str):
                elts.append(v.s)
            elif isinstance(v, ast.Constant):
                elts.append(v.value)
            else:
                return None
        return tuple(elts)  # type: ignore

    negate = False
    if isinstance(slc, ast.UnaryOp) and isinstance(slc.op, ast.USub):
        negate = True
        slc = slc.operand

    if isinstance(slc, ast.Name):
        return slc

    if not isinstance(slc, (ast.Constant, ast.Str, ast.Num)):
        return None

    if isinstance(slc, ast.Constant):
        slc = slc.value
    elif isinstance(slc, ast.Num):  # pragma: no cover
        slc = slc.n  # type: ignore
        if not isinstance(slc, int):
            return None
    elif isinstance(slc, ast.Str):  # pragma: no cover
        slc = slc.s  # type: ignore
    else:
        return None

    if isinstance(slc, int) and negate:
        slc = -slc  # type: ignore
    return slc  # type: ignore


class Atom(CommonEqualityMixin):
    def __init__(
        self,
        value: SupportedIndexType,
        is_callpoint: bool = False,
        is_subscript: bool = False,
        is_reactive: bool = False,
        is_blocking: bool = False,
    ):
        self.value = value
        self.is_callpoint = is_callpoint
        self.is_subscript = is_subscript
        self.is_reactive = is_reactive
        self.is_blocking = is_blocking

    def nonreactive(self) -> "Atom":
        return self.__class__(
            self.value,
            is_callpoint=self.is_callpoint,
            is_subscript=self.is_subscript,
            is_reactive=False,
        )

    def reactive(self) -> "Atom":
        return self.__class__(
            self.value,
            is_callpoint=self.is_callpoint,
            is_subscript=self.is_subscript,
            is_reactive=True,
        )

    def blocking(self) -> "Atom":
        return self.__class__(
            self.value,
            is_callpoint=self.is_callpoint,
            is_subscript=self.is_subscript,
            is_reactive=False,
            is_blocking=True,
        )

    def __hash__(self):
        return hash(
            (
                self.value,
                self.is_callpoint,
                self.is_subscript,
                self.is_reactive,
                self.is_blocking,
            )
        )

    def __repr__(self):
        return repr(str(self))

    def __str__(self):
        return self.value + ("(...)" if self.is_callpoint else "")


class SymbolRefVisitor(ast.NodeVisitor):
    def __init__(self):
        self.symbol_chain: List[Atom] = []

    def __call__(
        self,
        node: Union[
            ast.Attribute,
            ast.Subscript,
            ast.Call,
            ast.Name,
            ast.ClassDef,
            ast.FunctionDef,
            ast.AsyncFunctionDef,
        ],
    ) -> "SymbolRef":
        self.visit(node)
        self.symbol_chain.reverse()
        ret = SymbolRef(self.symbol_chain)
        self.symbol_chain = []
        return ret

    def _append_atom(
        self,
        node: Union[
            ast.Name, ast.Attribute, ast.ClassDef, ast.FunctionDef, ast.AsyncFunctionDef
        ],
        val: str,
        **kwargs,
    ) -> None:
        self.symbol_chain.append(
            Atom(
                val,
                is_reactive=id(node) in tracer().reactive_node_ids,
                is_blocking=id(node) in tracer().blocking_node_ids,
                **kwargs,
            )
        )

    def visit_Call(self, node):
        if isinstance(node.func, ast.Attribute):
            self._append_atom(node.func, node.func.attr, is_callpoint=True)
            self.visit(node.func.value)
        elif isinstance(node.func, ast.Subscript):
            if isinstance(node.func.slice, ast.Constant):
                self.symbol_chain.append(
                    Atom(str(node.func.slice.value), is_callpoint=True)
                )
            elif isinstance(node.func.slice, ast.Index) and isinstance(
                node.func.slice.value, (ast.Str, ast.Num)
            ):
                if isinstance(node.func.slice.value, ast.Str):
                    self.symbol_chain.append(
                        Atom(
                            node.func.slice.value.s,
                            is_callpoint=True,
                            is_subscript=True,
                        )
                    )
                else:
                    self.symbol_chain.append(
                        Atom(
                            str(node.func.slice.value.n),
                            is_callpoint=True,
                            is_subscript=True,
                        )
                    )
                self.visit(node.func.value)
        elif isinstance(node.func, ast.Name):
            self.visit(node.func)
            self.symbol_chain[-1].is_callpoint = True
        elif isinstance(node.func, ast.Call):
            # TODO: handle this case too, e.g. f.g()().h
            pass
        elif isinstance(node.func, ast.Lambda):
            # TODO: handle this case too, e.g. (lambda: [1, 2])()
            pass
        else:
            raise TypeError("invalid type for node.func %s" % node.func)

    def visit_Attribute(self, node: ast.Attribute) -> None:
        self._append_atom(node, node.attr)
        self.visit(node.value)

    def visit_Subscript(self, node: ast.Subscript) -> None:
        resolved = resolve_slice_to_constant(node)
        if resolved is not None:
            if isinstance(resolved, ast.Name):
                # FIXME: hack to make the static checker stop here
                # In the future, it should try to attempt to resolve
                # the value of the ast.Name node
                pass
            else:
                self.symbol_chain.append(Atom(resolved, is_subscript=True))
        self.visit(node.value)

    def visit_Name(self, node: ast.Name) -> None:
        self._append_atom(node, node.id)

    def visit_ClassDef(self, node: ast.ClassDef) -> None:
        self._append_atom(node, node.name)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> None:
        self._append_atom(node, node.name)

    def visit_AsyncFunctionDef(self, node: ast.AsyncFunctionDef) -> None:
        self._append_atom(node, node.name)

    def generic_visit(self, node):
        # raise ValueError('we should never get here: %s' % node)
        # give up
        return


class SymbolRef(CommonEqualityMixin):
    _cached_symbol_ref_visitor = SymbolRefVisitor()

    def __init__(self, symbols: Union[ast.AST, Atom, Sequence[Atom]]):
        # FIXME: each symbol should distinguish between attribute and subscript
        # FIXME: bumped in priority 2021/09/07
        if isinstance(
            symbols,
            (
                ast.Name,
                ast.Attribute,
                ast.Subscript,
                ast.Call,
                ast.ClassDef,
                ast.FunctionDef,
                ast.AsyncFunctionDef,
            ),
        ):
            symbols = self._cached_symbol_ref_visitor(symbols).chain
        elif isinstance(symbols, ast.AST):  # pragma: no cover
            raise TypeError("unexpected type for %s" % symbols)
        elif isinstance(symbols, Atom):
            symbols = [symbols]
        self.chain: Tuple[Atom, ...] = tuple(symbols)

    def __hash__(self):
        return hash(self.chain)

    def __repr__(self):
        return repr(self.chain)

    def __str__(self):
        return repr(self)

    def nonreactive(self) -> "SymbolRef":
        return self.__class__([atom.nonreactive() for atom in self.chain])

    def gen_resolved_symbols(
        self,
        scope: "Scope",
        only_yield_final_symbol: bool,
        yield_all_intermediate_symbols: bool = False,
        inherit_reactivity: bool = True,
        yield_in_reverse: bool = False,
    ) -> Generator[ResolvedDataSymbol, None, None]:
        assert not (only_yield_final_symbol and yield_all_intermediate_symbols)
        assert not (yield_in_reverse and not yield_all_intermediate_symbols)
        dsym, atom, next_atom = None, None, None
        reactive_seen = False
        blocking_seen = False
        if yield_in_reverse:
            gen: Iterable[Tuple["DataSymbol", Atom, Atom]] = [
                (resolved.dsym, resolved.atom, resolved.next_atom)
                for resolved in self.gen_resolved_symbols(
                    scope,
                    only_yield_final_symbol=only_yield_final_symbol,
                    yield_all_intermediate_symbols=True,
                    inherit_reactivity=False,
                    yield_in_reverse=False,
                )
            ]
            cast(list, gen).reverse()
        else:
            gen = scope.gen_data_symbols_for_attrsub_chain(self)
        for dsym, atom, next_atom in gen:
            reactive_seen = reactive_seen or atom.is_reactive
            yield_all_intermediate_symbols = (
                yield_all_intermediate_symbols or reactive_seen
            )
            if inherit_reactivity:
                if reactive_seen and not blocking_seen and not atom.is_reactive:
                    atom = atom.reactive()
                elif blocking_seen and not atom.is_blocking:
                    atom = atom.blocking()
            if yield_all_intermediate_symbols:
                # TODO: only use this branch one staleness checker can be smarter about liveness timestamps.
                #  Right now, yielding the intermediate elts of the chain will yield false positives in the
                #  event of namespace stale children.
                yield ResolvedDataSymbol(dsym, atom, next_atom)
        if not yield_all_intermediate_symbols and dsym is not None:
            if next_atom is None or not only_yield_final_symbol:
                yield ResolvedDataSymbol(dsym, atom, next_atom)


class LiveSymbolRef(CommonEqualityMixin):
    def __init__(self, ref: SymbolRef, timestamp: int) -> None:
        self.ref = ref
        self.timestamp = timestamp

    def __hash__(self):
        return hash((self.ref, self.timestamp))

    def gen_resolved_symbols(
        self,
        scope: "Scope",
        only_yield_final_symbol: bool,
        yield_all_intermediate_symbols: bool = False,
    ):
        blocking_seen = False
        for resolved_sym in self.ref.gen_resolved_symbols(
            scope,
            only_yield_final_symbol,
            yield_all_intermediate_symbols=yield_all_intermediate_symbols,
        ):
            resolved_sym.liveness_timestamp = self.timestamp
            blocking_seen = blocking_seen or resolved_sym.is_blocking
            if blocking_seen and not resolved_sym.is_blocking:
                resolved_sym.atom = resolved_sym.atom.blocking()
            yield resolved_sym

    def __str__(self):
        return str(self.ref)



# -*- coding: utf-8 -*-
import logging
from typing import TYPE_CHECKING, Iterable, NamedTuple, Optional, Union

from nbsafety.singletons import nbs, tracer

if TYPE_CHECKING:
    # avoid circular imports
    from nbsafety.data_model.data_symbol import DataSymbol


logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)


class Timestamp(NamedTuple):
    cell_num: int
    stmt_num: int

    @classmethod
    def current(cls) -> "Timestamp":
        # TODO: shouldn't have to go through nbs() singleton to get the cell counter,
        #  but the dependency structure prevents us from importing from nbsafety.data_model.code_cell
        return cls(nbs().cell_counter(), tracer().module_stmt_counter())

    @classmethod
    def uninitialized(cls) -> "Timestamp":
        return cls(-1, -1)

    @property
    def is_initialized(self):
        return self > Timestamp.uninitialized()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Timestamp):
            raise TypeError(
                "cannot compare non-timestamp value %s with timestamp %s"
                % (other, self)
            )
        return tuple(self._asdict().values()) == tuple(other._asdict().values())

    @classmethod
    def update_usage_info(
        cls,
        symbols: Union[Optional["DataSymbol"], Iterable[Optional["DataSymbol"]]],
        exclude_ns=False,
    ):
        if symbols is None:
            return
        try:
            iter(symbols)  # type: ignore
        except TypeError:
            symbols = [symbols]  # type: ignore
        used_time = cls.current()
        for sym in symbols:  # type: ignore
            if sym is not None and not sym.is_anonymous:
                sym.update_usage_info(used_time=used_time, exclude_ns=exclude_ns)



# -*- coding: utf-8 -*-
import asyncio
import inspect
import sys

from ipykernel.ipkernel import IPythonKernel

from nbsafety.version import __version__
from nbsafety.safety import NotebookSafety
from nbsafety.singletons import nbs


class SafeKernel(IPythonKernel):
    implementation = "kernel"
    implementation_version = __version__

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        NotebookSafety.instance(use_comm=True)
        import nest_asyncio

        # ref: https://github.com/erdewit/nest_asyncio
        nest_asyncio.apply()

    def init_metadata(self, parent):
        """
        Don't actually change the metadata; we just want to get the cell id
        out of the execution request.
        """
        metadata = parent.get("metadata", {})
        cell_id = metadata.get("cellId", None)
        if cell_id is not None:
            nbs().set_active_cell(cell_id)
        tags = tuple(metadata.get("tags", ()))
        nbs().set_tags(tags)
        return super().init_metadata(parent)

    if inspect.iscoroutinefunction(IPythonKernel.do_execute):

        async def do_execute(
            self,
            code,
            silent,
            store_history=False,
            user_expressions=None,
            allow_stdin=False,
        ):
            super_ = super()

            async def _run_cell_func(cell):
                return await super_.do_execute(
                    cell, silent, store_history, user_expressions, allow_stdin
                )

            if silent:
                # then it's probably a control message; don't run through nbsafety
                return await _run_cell_func(code)
            else:
                return await nbs().safe_execute(code, True, _run_cell_func)

    else:

        def do_execute(
            self,
            code,
            silent,
            store_history=False,
            user_expressions=None,
            allow_stdin=False,
        ):
            super_ = super()

            def _run_cell_func(cell):
                return super_.do_execute(
                    cell, silent, store_history, user_expressions, allow_stdin
                )

            return next(
                iter(
                    asyncio.get_event_loop().run_until_complete(
                        asyncio.wait([nbs().safe_execute(code, False, _run_cell_func)])
                    )[0]
                )
            ).result()



# -*- coding: utf-8 -*-
from typing import TYPE_CHECKING

import pyccolo as pyc
from traitlets.config.configurable import SingletonConfigurable

if TYPE_CHECKING:
    from nbsafety.safety import NotebookSafety as NotebookSafetyInstance
    from nbsafety.tracing.nbsafety_tracer import SafetyTracer as TracerInstance


class NotebookSafety(SingletonConfigurable):
    _Xyud34_INSTANCE = None

    def __init__(self):
        super().__init__()
        # we need to keep another ref around for some reason to prevent a segfault
        # TODO: figure out why
        self.__class__._Xyud34_INSTANCE = self


class SingletonBaseTracer(pyc.BaseTracer):
    pass


def nbs() -> "NotebookSafetyInstance":
    assert NotebookSafety.initialized()
    return NotebookSafety.instance()


def tracer() -> "TracerInstance":
    assert SingletonBaseTracer.initialized()
    return SingletonBaseTracer.instance()



# -*- coding: utf-8 -*-
import ast
import astunparse
import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Dict, List, Set, Type

from nbsafety.data_model.timestamp import Timestamp
from nbsafety.singletons import nbs
from nbsafety.types import TimestampOrCounter

if TYPE_CHECKING:
    from nbsafety.data_model.code_cell import ExecutedCodeCell


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


def _get_ts_dependencies(
    timestamp: TimestampOrCounter,
    dependencies: Set[TimestampOrCounter],
    timestamp_to_ts_deps: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
) -> None:
    """
    For a given timestamps, this function recursively populates a set of
    timestamps that the given timestamp depends on, based on the live symbols.

    Args:
        - timestamp (ts_or_int): current timestamp / cell to get dependencies for
        - dependencies (set<ts_or_int>): set of timestamps / cell coutners so far that exist
        - cell_num_to_dynamic_deps (dict<ts_or_int, set<ts_or_int>>): mapping from used timestamp
        to timestamp of symbol definition
        - timestamp_to_static_ts_deps (dict<ts_or_int, set<ts_or_int>>): mapping from used timestamp
        to timestamp of symbol definition (for statically computed timestamps)

    Returns:
        None
    """
    # Base case: cell already in dependencies
    if timestamp in dependencies:
        return
    if isinstance(timestamp, int) and timestamp < 0:
        return
    if isinstance(timestamp, Timestamp) and not timestamp.is_initialized:
        return

    # Add current cell to dependencies
    dependencies.add(timestamp)

    # Retrieve cell numbers for the dependent symbols
    # Add dynamic and static dependencies
    dep_timestamps = timestamp_to_ts_deps[timestamp]
    logger.info("dynamic ts deps for %s: %s", timestamp, dep_timestamps)

    # For each dependent cell, recursively get their dependencies
    for ts in dep_timestamps - dependencies:
        _get_ts_dependencies(ts, dependencies, timestamp_to_ts_deps)


def _coarsen_timestamps(graph: Dict[Timestamp, Set[Timestamp]]) -> Dict[int, Set[int]]:
    coarsened: Dict[int, Set[int]] = defaultdict(set)
    for child, parents in graph.items():
        for par in parents:
            coarsened[child.cell_num].add(par.cell_num)
    return coarsened


def _graph_union(
    graph: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
    subsumed: Dict[TimestampOrCounter, Set[TimestampOrCounter]],
):
    for child, parents in subsumed.items():
        graph[child] |= parents
    return graph


def _compute_slice_impl(seeds: List[TimestampOrCounter]) -> Set[TimestampOrCounter]:
    assert len(seeds) > 0
    dependencies: Set[TimestampOrCounter] = set()
    timestamp_to_ts_deps: Dict[
        TimestampOrCounter, Set[TimestampOrCounter]
    ] = defaultdict(set)
    if nbs().mut_settings.dynamic_slicing_enabled:
        if isinstance(seeds[0], Timestamp):
            timestamp_to_ts_deps = _graph_union(
                timestamp_to_ts_deps, nbs().dynamic_data_deps
            )
        else:
            timestamp_to_ts_deps = _graph_union(
                timestamp_to_ts_deps, _coarsen_timestamps(nbs().dynamic_data_deps)
            )
    if nbs().mut_settings.static_slicing_enabled:
        if isinstance(seeds[0], Timestamp):
            timestamp_to_ts_deps = _graph_union(
                timestamp_to_ts_deps, nbs().static_data_deps
            )
        else:
            timestamp_to_ts_deps = _graph_union(
                timestamp_to_ts_deps, _coarsen_timestamps(nbs().static_data_deps)
            )

    # ensure we at least get the static deps
    for seed in seeds:
        _get_ts_dependencies(seed, dependencies, timestamp_to_ts_deps)
    if isinstance(seeds[0], Timestamp):
        for seed in seeds:
            for ts in list(timestamp_to_ts_deps.keys()):
                if ts.cell_num == seed.cell_num:
                    _get_ts_dependencies(ts, dependencies, timestamp_to_ts_deps)
    return dependencies


class CodeCellSlicingMixin:
    def compute_slice(  # type: ignore
        self: "ExecutedCodeCell", stmt_level: bool = False
    ) -> Dict[int, str]:
        return self.compute_slice_for_cells({self}, stmt_level=stmt_level)

    @classmethod
    def compute_slice_for_cells(  # type: ignore
        cls: Type["ExecutedCodeCell"],
        cells: Set["ExecutedCodeCell"],
        stmt_level: bool = False,
    ) -> Dict[int, str]:
        """
        Gets a dictionary object of cell dependencies for the cell with
        the specified execution counter.

        Args:
            - cell_num (int): cell to get dependencies for, defaults to last
                execution counter

        Returns:
            - dict (int, str): map from required cell number to code
                representing dependencies
        """
        if stmt_level:
            stmts_by_cell_num = cls.compute_slice_stmts_for_cells(cells)
            for cell in cells:
                stmts_by_cell_num.pop(cell.cell_ctr, None)
            ret = {
                ctr: "\n".join(astunparse.unparse(stmt).strip() for stmt in stmts)
                for ctr, stmts in stmts_by_cell_num.items()
            }
            for cell in cells:
                ret[cell.cell_ctr] = cell.content
            return ret
        else:
            deps: Set[int] = _compute_slice_impl([cell.cell_ctr for cell in cells])
            return {dep: cls.from_timestamp(dep).content for dep in deps}

    def compute_slice_stmts(  # type: ignore
        self: "ExecutedCodeCell",
    ) -> Dict[int, List[ast.stmt]]:
        return self.compute_slice_stmts_for_cells({self})

    @classmethod
    def compute_slice_stmts_for_cells(  # type: ignore
        cls: Type["ExecutedCodeCell"],
        cells: Set["ExecutedCodeCell"],
    ) -> Dict[int, List[ast.stmt]]:
        deps_stmt: Set[Timestamp] = _compute_slice_impl(
            [Timestamp(cell.cell_ctr, -1) for cell in cells]
        )
        stmts_by_cell_num = defaultdict(list)
        seen_stmt_ids = set()
        for ts in sorted(deps_stmt):
            stmt = cls.from_timestamp(ts.cell_num).to_ast().body[ts.stmt_num]
            stmt_id = id(stmt)
            if stmt is None or stmt_id in seen_stmt_ids:
                continue
            seen_stmt_ids.add(stmt_id)
            if stmt is not None:
                stmts_by_cell_num[ts.cell_num].append(stmt)
        for cell in cells:
            stmts_by_cell_num[cell.cell_ctr] = list(cell.to_ast().body)
        return dict(stmts_by_cell_num)
